import cPickle, zlib

from twisted.internet.defer import inlineCallbacks, DeferredList, returnValue
from twisted.internet import reactor
from twisted.python import log

from txyam.utils import ketama, deferredDict
from txyam.factory import MemCacheClientFactory


class NoServerError(Exception):
    """
    No available connected servers to accept command.
    """

class InvalidHostPortError(Exception):
    """
    Invalid host/port specification.
    """


def wrap(cmd):
    """
    Used to wrap all of the memcache methods (get,set,getMultiple,etc).
    """
    def wrapper(self, key, *args, **kwargs):
        func = getattr(self.getClient(key), cmd)
        return func(key, *args, **kwargs)
    return wrapper


class YamClient:
    def __init__(self, hosts, connect=True):
        """
        @param hosts: A C{list} of C{tuple}s containing hosts and ports.
        """
        self.hosts = hosts
        if connect:
            self.connect()


    def getActiveConnections(self):
        return [factory.client for factory in self.factories if not factory.client is None]


    def getClient(self, key):
        hosts = self.getActiveConnections()
        log.msg("Using %i active hosts" % len(hosts))
        if len(hosts) == 0:
            raise NoServerError, "No connected servers remaining."
        return hosts[ketama(key) % len(hosts)]
    

    @inlineCallbacks
    def connect(self):
        self.factories = []
        for hp in self.hosts:
            if isinstance(hp, tuple):
                host, port = hp
            elif isinstance(hp, str):
                host = hp
                port = 11211
            else:
                raise InvalidHostPortError, "Connection info should be either hostnames or host/port tuples"
            factory = MemCacheClientFactory()
            connection = yield reactor.connectTCP(host, port, factory)
            self.factories.append(factory)

        # fire callback when all connections have been established
        yield DeferredList([factory.deferred for factory in self.factories])
        returnValue(self)


    def disconnect(self):
        log.msg("Disconnecting from all clients.")
        for factory in self.factories:
            factory.stopTrying()
        for connection in self.getActiveConnections():
            connection.transport.loseConnection()            


    def flushAll(self):
        hosts = self.getActiveConnections()        
        log.msg("Flushing %i hosts" % len(hosts))
        return DeferredList([host.flushAll() for host in hosts])


    def stats(self, arg=None):
        ds = {}
        for factory in self.factories:
            if not factory.client is None:
                hp = "%s:%i" % (factory.addr.host, factory.addr.port)
                ds[hp] = factory.client.stats(arg)
        log.msg("Getting stats on %i hosts" % len(ds))
        return deferredDict(ds)


    def version(self):
        ds = {}
        for factory in self.factories:
            if not factory.client is None:
                hp = "%s:%i" % (factory.addr.host, factory.addr.port)
                ds[hp] = factory.client.version()
        log.msg("Getting version on %i hosts" % len(ds))
        return deferredDict(ds)


    def pickle(self, value, compress):
        p = cPickle.dumps(value, cPickle.HIGHEST_PROTOCOL)
        if compress:
            p = zlib.compress(p)
        return p


    def unpickle(self, value, uncompress):
        if uncompress:
            value = zlib.decompress(value)
        return cPickle.loads(value)            


    def setPickled(self, key, value, **kwargs):
        value = self.pickle(value, kwargs.pop('compress', False))
        return self.set(key, value, **kwargs)


    def addPickled(self, key, value, **kwargs):
        value = self.pickle(value, kwargs.pop('compress', False))
        return self.add(key, value, **kwargs)


    def getPickled(self, key, **kwargs):
        def handleResult(result, uncompress):
            index = len(result) - 1
            result = list(result)
            if result[index] is not None:
                result[index] = self.unpickle(result[index], uncompress)
            return tuple(result)
        uncompress = kwargs.pop('uncompress', False)
        return self.get(key, **kwargs).addCallback(handleResult, uncompress)


    # Following methods can be found at
    # http://twistedmatrix.com/trac/browser/tags/releases/twisted-12.0.0/twisted/protocols/memcache.py
    set = wrap("set")
    get = wrap("get")
    increment = wrap("increment")
    decrement = wrap("decrement")
    replace = wrap("replace")
    add = wrap("add")
    set = wrap("set")
    checkAndSet = wrap("checkAndSet")
    append = wrap("append")
    prepend = wrap("prepend")
    getMultiple = wrap("getMultiple")
    delete = wrap("delete")


def ConnectedYamClient(hosts):
    return YamClient(hosts, connect=False).connect()



from twisted.internet import reactor, protocol
from twisted.protocols.memcache import MemCacheProtocol
from twisted.internet.defer import inlineCallbacks

from txyam.client import YamClient

class YamClientWrapper(YamClient):
    def __init__(self):
        pass

    
    def setConnection(self, host, port):
        def save(connection):
            self.connection = connection
            return self
        
        clientCreator = protocol.ClientCreator(reactor, MemCacheProtocol)        
        return clientCreator.connectTCP(host, port).addCallback(save)


    def getClient(self, key):
        # we only have one connection
        return self.connection


    def getActiveConnections(self):
        # we only have one connection
        return [self.connection]


    def disconnect(self):
        return self.connection.transport.loseConnection()



from twisted.trial import unittest
from twisted.internet.defer import inlineCallbacks

from utils import YamClientWrapper

class FactoryTest(unittest.TestCase):

    @inlineCallbacks
    def setUp(self):
        self.client = YamClientWrapper()
        yield self.client.setConnection('localhost', 11211)


    @inlineCallbacks
    def test_connection(self):
        yield self.client.set("testkey", '123')
        result = yield self.client.get("testkey")
        self.assertEqual(result[1], '123')
        

    def tearDown(self):
        return self.client.disconnect()



from twisted.internet.defer import Deferred
from twisted.internet.protocol import ReconnectingClientFactory
from twisted.python import log
from twisted.protocols.memcache import MemCacheProtocol

class ConnectingMemCacheProtocol(MemCacheProtocol):
    def connectionMade(self):
        self.factory.connectionMade()


class MemCacheClientFactory(ReconnectingClientFactory):
    initialDelay = 0.1
    protocol = ConnectingMemCacheProtocol
    noisy = True

    def __init__(self):
        self.client = None
        self.addr = None
        self.deferred = Deferred()


    def buildProtocol(self, addr):
        self.client = self.protocol()
        self.addr = addr
        self.client.factory = self
        self.resetDelay()
        return self.client


    def clientConnectionLost(self, connector, reason):
        log.msg("Lost connection to %s - %s" % (self.addr, reason))
        self.client = None
        ReconnectingClientFactory.clientConnectionLost(self, connector, reason)


    def clientConnectionFailed(self, connector, reason):
        log.msg("Connection failed to %s - %s" % (self.addr, reason))
        self.client = None
        ReconnectingClientFactory.clientConnectionFailed(self, connector, reason)


    def connectionMade(self):
        self.deferred.callback(self)



#!/usr/bin/env python
from setuptools import setup, find_packages

setup(
    name="txyam",
    version="0.2",
    description="Yet Another Memcached (YAM) client for Twisted.",
    author="Brian Muller",
    author_email="bamuller@gmail.com",
    license="MIT",
    url="http://github.com/bmuller/txyam",
    packages=find_packages(),
    requires=["twisted.protocols.memcache"],
    install_requires=['twisted>=12.0']
)



import hashlib

from twisted.internet import defer

def signedInt(n):
    n = n & 0xffffffff
    return ((1 << 31) & n) and (~n + 1) or n


def ketama(key):
    """
    MD5-based hashing algorithm used in consistent hashing scheme
    to compensate for servers added/removed from memcached pool.

    Copied from dpnova / txmemcache found at:
    https://github.com/dpnova/txmemcache/blob/master/txmemcache/hash.py
    """
    d = hashlib.md5(key).digest()
    c = signedInt
    h = c((ord(d[3])&0xff) << 24) | c((ord(d[2]) & 0xff) << 16) | c((ord(d[1]) & 0xff) << 8) | c(ord(d[0]) & 0xff)
    return h


def deferredDict(d):
    """
    Just like a C{defer.DeferredList} but instead accepts and returns a C{dict}.

    @param d: A C{dict} whose values are all C{Deferred} objects.

    @return: A C{DeferredList} whose callback will be given a dictionary whose
    keys are the same as the parameter C{d}'s and whose values are the results
    of each individual deferred call.
    """
    if len(d) == 0:
        return defer.succeed({})

    def handle(results, names):
        rvalue = {}
        for index in range(len(results)):
            rvalue[names[index]] = results[index][1]
        return rvalue

    dl = defer.DeferredList(d.values())
    return dl.addCallback(handle, d.keys())
